<?php
//
// Step 1 : Build JSON map for every object of the default pack
//

$basemapFile = __DIR__ . '/base-map.dungeondraft_map';
$pckFile=($argv[1] ?? "/opt/Dungeondraft/Dungeondraft.pck");
$pckName = basename($pckFile, '.pck');
$baseDir = __DIR__ . "/$pckName";
if (!is_dir($baseDir)) {
	echo "Create $baseDir/... directories\n";
	mkdir($baseDir);
	mkdir("$baseDir/assets-png");
	mkdir("$baseDir/snapshots");
	$gridIncrement = 5;
} else {
	$gridIncrement = 1;
}

echo "Reading the pack...\n";
$map = file_get_contents(__DIR__ . '/base-map.dungeondraft_map');
$pack = file_get_contents($pckFile);
if (!preg_match_all('~(res://textures/objects/[a-z0-9/A-Z_-]+\\.png)(?!\\.import)~',$pack,$matches))
	throw new LogicException("Failed to find object resources in $pckFile");

if (!preg_match('~"width": ([0-9]+),~', $map, $baseWidth))
	throw new LogicException("Missing width in base map");
if (!preg_match('~"height": ([0-9]+),~', $map, $baseHeight))
	throw new LogicException("Missing height in base map");

$mapWidth = $baseWidth[1];
$mapHeight = $baseHeight[1];
echo "Base map is w={$mapWidth} x h={$mapHeight[1]}\n";

// Let's make this static
$tagNextId = "\"next_node_id\": \"HERE\",";
$tagObjects = "\t\t\t\t\"objects\": [\n\n\t\t\t\t";
$map = str_replace(str_replace('HERE', '0', $tagNextId), str_replace('HERE', 'ffff', $tagNextId), $map);
$posObject = strpos($map, $tagObjects);
$mapBegin = substr($map, 0, $posObject) . "\n\"objects\": ";
$mapEnd = substr($map, $posObject + strlen($tagObjects) + 1);
$genMapPath = "$baseDir/gen-map.dungeondraft_map";
$resRemain = array_unique(array_values($matches[1]));
$mapFactor = 256;

echo "Total unique items: " . count($resRemain) . "\n";
$gridSize = $gridIncrement;
while (true) {
	$x = 0;
	$y = $gridSize;
	$objects = array();
	$places = array();
	
	//
	// Create a map containing items to export
	//
	foreach ($resRemain as $i => $res) {
		$basename = basename(str_replace(array('res://textures/objects/', '/'), array('', '-'), $res), '.png');
		if (is_file("$baseDir/assets-png/$basename.png")) {
//			echo "PNG $basename already exists: skip\n";
			continue;
		}
		
		$x += $gridSize;
		if ($x > ($mapWidth - $gridSize)) {
			$x = $gridSize;
			$y += $gridSize;
			if ($y > ($mapHeight - $gridSize)) {
				echo "\033[33mMap is full, proceeding to export\033[0m\n";
				break;
			}
		}
		
//		echo "Adding $res to map\n";
		$places[] = array($i, $x, $y, $basename, $res);
		$xx = $x * $mapFactor;
		$yy = $y * $mapFactor;
		$object = array(
			"position"=> "Vector2( $xx, $yy )",
			"rotation"=> 0,
			"scale"=> "Vector2( 1, 1 )",
			"mirror"=> false,
			"texture"=> $res,
			"layer"=> 100,
			"shadow"=> false,
			"custom_color" => "ffff0000",
			"block_light"=> false,
			"node_id"=> "1");
		$objects[] = $object;
	}

	if (!$objects)
		break; // All done!
	
	echo "\033[34mItems added to the map: " . count($objects) . "\033[0m\n";
	$jsonObjects = str_replace('\\', '', json_encode($objects, JSON_PRETTY_PRINT));
	file_put_contents($genMapPath, $mapBegin . $jsonObjects . $mapEnd);
	
	echo "Opening DD so you can export the map as PNG in " . __DIR__ . "\n";
	
	//
	// Ask you to do the export
	// 
	while (true) {
		$pngs = glob("$baseDir/*.png");
		if (count($pngs) === 0)
			break;
		echo "\033[33mThere must be no PNG file in $baseDir, found " . count($pngs) . "\033[0m\n";
		sleep(1);
	}
	exec("/opt/Dungeondraft/Dungeondraft.x86_64 \"$genMapPath\"");
	
	while (true) {
		$pngs = glob("$baseDir/*.png");
		if (count($pngs) === 1)
			break;
		echo "\033[33mThere must be 1 single PNG file in $baseDir, found " . count($pngs) . "\033[0m\n";
		sleep(1);
	}
	$png = $pngs[0];
	
	//
	// Get PNG's metadata
	//
	echo "Scanning content of $png\n";
	$img = imagecreatefrompng($png);
	$dpi = imagesx($img) / $mapWidth;
	if (!is_int($dpi))
		throw new LogicException("DPI of the map PNG is $dpi which is not integer");
	$dpiy = (imagesy($img) / $mapHeight);
	if ($dpi !== $dpiy)
		throw new LogicException("DPI for x ($dpi) does not match dpi for y ($dpiy)");
	
	echo "Image DPI is $dpi\n";
	
	//
	// Isolate items in that PNG
	//
	$gdpi = $gridSize * $dpi;
	$exporteds = array();
	foreach ($places as list($i, $x, $y, $exportName, $res)) {
//		echo "Trying to isolate $exportName...\n";
		$isolated = imagecreatetruecolor($gdpi, $gdpi);
		imagealphablending($isolated, false);
		imagesavealpha($isolated, true);
		imagefill($isolated, 0, 0, 0x7f000000);
		imagecopy($isolated, $img, 0, 0, (int)(($x-$gridSize/2)*$dpi), (int)(($y-$gridSize/2)*$dpi), $gdpi, $gdpi);
		
		// Crop closest
		$crops = array('x' => 0, 'y' => 0, 'width' => $gdpi, 'height' => $gdpi);
		$cropsValid = 0;
		for ($x0=0; $x0 < $gdpi; $x0++) {
			for ($y0=0; $y0 < $gdpi; $y0++) {
				if (imagecolorat($isolated, $x0, $y0) !== 0x7f000000)
					break 2;
			}
			$crops['x']++;
			$crops['width']--;
		}
		for ($y0=0; $y0 < $gdpi; $y0++) {
			for ($x0=0; $x0 < $gdpi; $x0++) {
				if (imagecolorat($isolated, $x0, $y0) !== 0x7f000000)
					break 2;
			}
			$crops['y']++;
			$crops['height']--;
		}
		for ($x0=$gdpi-1; $x0 >= 0; $x0--) {
			for ($y0=0; $y0 < $gdpi; $y0++) {
				if (imagecolorat($isolated, $x0, $y0) !== 0x7f000000)
					break 2;
			}
			$crops['width']--;
		}
		for ($y0=$gdpi-1; $y0 >= 0; $y0--) {
			for ($x0=0; $x0 < $gdpi; $x0++) {
				if (imagecolorat($isolated, $x0, $y0) !== 0x7f000000)
					break 2;
			}
			$crops['height']--;
		}
		
		// Keep 1 pixel around for safety
		$crops['x']--;
		$crops['y']--;
		$crops['width']+=2;
		$crops['height']+=2;
		
		if ($crops['x'] <= 0 
			|| $crops['y'] <= 0 
			|| ($crops['x'] + $crops['width']) >= $gdpi
			|| ($crops['y'] + $crops['height']) >= $gdpi) {
			echo "The item $exportName is not isolated enough (" . implode(' ', $crops)
				 . "), it's probably too big or has a too big neighbor, skip\n";
			$exporteds[$i] = 0;
			$resRemain[] = $resRemain[$i];
			unset($resRemain[$i]);
			continue;
		}
		
		echo "Saving item to $exportName\n";
		$isolated = imagecrop($isolated, $crops);
		imagesavealpha($isolated, true);
		imagepng($isolated, "$baseDir/assets-png/$exportName.png");
		$exporteds[$i] = 1;
	}
	
	//
	// Prepare next iteration
	//
	$exportCount = array_sum($exporteds);
	$totalTries = count($exporteds);
	echo "Items exported: $exportCount / $totalTries\n";
	
	$exportPc = round($exportCount / $totalTries * 100, 2);
	if ($exportPc < (5 + $gridSize)) {
		echo "\033[34mI could only export $exportPc % of the objects, the grid seems too small\033[0m\n";
		$gridSize += $gridIncrement;
		echo "So increase the grid size and make it now $gridSize\n";
	}
	rename($png, "$baseDir/snapshots/" . date('Y-m-d-H-i-s') . ".png");
}
echo "Done, all PNG are in ./assets-png/*.png\n";

//
// Now, what elements are colorables?
//
require_once __DIR__ . '/rgb2hsl.php';
require_once __DIR__ . '/hsl2rgb.php';
function col2hsla(int $color): array {
	$a = $color >> 24;
	list($h,$s,$l) = rgb2hsl(array(($color >> 16)&0xff,($color >> 8)&0xff,($color >> 0)&0xff));
	$hsla = array((int)round($h*360),(int)round($s*100),(int)round($l*100),(int)$a);
//	echo "Color: " . dechex($color) . "\tHSLA: " . implode(' ', $hsla) . "\n";
	return $hsla;
}
function hsla2col(int $h, int $s, int $l, int $a): int {
	list('r'=>$r,'g'=>$g,'b'=>$b) = hsl2rgb((($h+360)%360)/360, min(max($s,0),100)/100, min(max($l,0),100)/100);
	$color = ($a << 24) + (((int)round($r)) << 16) + (((int)round($g)) << 8) + (((int)round($b)) << 0);
//	echo "HSLA: $h $s $l $a\tColor: " . dechex($color) . "\n";
	return (int)$color;
}

$packDir = "$baseDir/pack-dd";
if (!is_dir($packDir))
	throw new DomainException("Please create a directory $packDir that contains the DD pack base (copy 'pack-model')");

$packName = "$pckName Colorables";
$packJson = array(
	"name" => $packName,
	"id" => substr(base64_encode(md5($packName)), 0, 8),
	"version" => date('1.Y.m.d.H.i.s'),
	"author" => "Xenos Script",
	"keywords" => "",
	"allow_3rd_party_mapping_software_to_read" => false,
	"custom_color_overrides" => array(
		"enabled" => false,
		"min_redness" => 0.1,
		"min_saturation" => 0,
		"red_tolerance" => 0.04));
file_put_contents("$packDir/pack.json", json_encode($packJson, JSON_PRETTY_PRINT));

if (!is_dir("$packDir/textures"))
	mkdir("$packDir/textures");
if (!is_dir("$packDir/textures/objects"))
	mkdir("$packDir/textures/objects");
if (!is_dir("$packDir/textures/objects/colorable"))
	mkdir("$packDir/textures/objects/colorable");
	
$objectsPath = "textures/objects/colorable/" . strtolower($pckName);
$colorablesDir = "$packDir/$objectsPath";
if (!is_dir($colorablesDir))
	mkdir($colorablesDir);

foreach (glob("$baseDir/assets-png/*.png") as $file) {
	$basename = basename($file, '.png');
	if (glob("$colorablesDir/$basename-*.png"))
		continue; // Image already in the pack, skip
	
	echo "Analyzing $file...\n";
	$colorHueSize = 60; // degrees of tolerence for color clustering
	$lightFactor = 1.0;
	$saturateFactor = 1.0;
	if (str_contains($file, 'cabbage'))
		$colorHueSize = 200;
	else if (str_contains($file, 'snow'))
		list($saturateFactor,$lightFactor) = array(1.5,0.6);
		
	$pixelsByHue = array();
	$img = imagecreatefrompng($file);
	for ($y = 0, $sy = imagesy($img); $y < $sy; $y++) {
	for ($x = 0, $sx = imagesx($img); $x < $sx; $x++) {
		$color = imagecolorat($img, $x, $y);
		list($h, $s, $l, $a) = col2hsla($color);
		if ($a >= 0x70)
			continue; // Transparent pixel, ignore
//		if ($s < 10 || $l < 10 || $l > 90) // Let's count them anyway
//			continue; // Pixel too grey, or too dark, or too white: ignore
		if (($h > 355 || $h < 5) && $s > 80 && $l > 40 && $l < 60) {
			echo "There must be no redish pixels, got one at $x $y: it's probably already a colorable, so skip\n";
			continue 3;
		}
		
		// Cluster by closest color
		if (!array_key_exists($h, $pixelsByHue))
			$pixelsByHue[$h] = array();
		$pixelsByHue[$h][] = array($x, $y, $h, $s, $l, $a);
	}
	}
	
	echo "Clustering the image pixels...\n";
	$minCluster = array_sum(array_map('count',$pixelsByHue)) / 100 * 10; // At least xx% of the pixels image must be in a cluster
	if ($minCluster < 20)
		continue; // There are not enough non-transparent pixels here
	
	$hues = range(0, 360-$colorHueSize, 1); // Start pos of each hue cluster
	while (true) {
		$bestHue = array($minCluster);
		foreach ($hues as $i => $hue) {
			$avgHue = ($hue + $colorHueSize/2);
			$cluster = array();
			$weight = 0;
			$elements = 0;
			foreach ($pixelsByHue as $h => $pixels) {
				if ($h >= $hue && $h <= ($hue + $colorHueSize)) {
					$cluster[] = $pixels;
					$weight += count($pixels) * pow(exp(-(($h - $avgHue)/($colorHueSize/2))**4),2);
					$elements += count($pixels);
				}
			}
			if ($elements < $minCluster)
				continue; // Not enough pixels in this cluster
			if ($weight < $bestHue[0])
				continue; // Not the best cluster
			$bestHue = array($weight, array_merge(...$cluster), $hue - $colorHueSize, $hue + $colorHueSize);
		}
		if (count($bestHue) === 1)
			break; // No more hue cluster, stop here
		
		// Create the PNG colored file
//		$avgHue = array_sum(array_column($bestHue[1], 2)) / count(array_column($bestHue[1], 2));
		$maxSat = max(array_column($bestHue[1], 3));
		$maxLight = max(array_column($bestHue[1], 4));
		if ($maxLight > 6) {
			// Export only if there are light-enough pixels; else it means it's just black block
			$png = imagecreatefrompng($file);
			imagealphablending($png, false);
			foreach ($bestHue[1] as list($x, $y, $h, $s, $l, $a)) {
	//			$newColor = hsla2col(round(($h - $avgHue)/$colorHueSize*5 + 360)%360,$s+floor(100-$maxSat),$l,$a);
				$newColor = hsla2col(0,($s+floor(100-$maxSat))*$saturateFactor,$l*$lightFactor,$a);
				imagesetpixel($png, $x, $y, $newColor);
			}
			
			$outname = "$basename-{$bestHue[2]}-{$bestHue[3]}.png";
			echo "Saving colorable image $outname\n";
			imagesavealpha($png, true);
			imagepng($png, "$colorablesDir/$outname");
		}
		// Remove hues
		foreach ($hues as $i => $hue) {
			if ($hue >= $bestHue[2] && $hue <= $bestHue[3])
				unset($hues[$i]); // This hue will no longer be considered in clustering
		}
	}
}

echo "Images saved, building tags...\n";
$packTags = array();
foreach (glob("$colorablesDir/*.png") as $file) {
	$basename = basename($file, '.png');
	if (!preg_match('~^(.+)-[^-]+--?[0-9]+-[0-9]+$~', $basename, $match))
		throw new LogicException("Invalid file name $basename in pack");
	$keywords = array_map('strtolower', explode('-', $match[1]));
	$keywords[] = 'colorable';
	$keywords[] = $pckName;
	
	foreach ($keywords as $key) {
		$keyword = strtoupper($key[0]) . substr($key, 1);
		if (!array_key_exists($keyword, $packTags))
			$packTags[$keyword] = array();
		$packTags[$keyword][] = "$objectsPath/$basename.png";
	}
}
$defaultJson = array(
	"tags" => $packTags,
	"sets" => array(
		"Colorables Set" => array_keys($packTags)));
file_put_contents("$packDir/data/default.dungeondraft_tags", str_replace('\\', '', json_encode($defaultJson, JSON_PRETTY_PRINT)));

echo "\033[32mAll saved, you may now pack the $packDir directory in DungeonDraft\033[0m\n";

